/*
======================================================================================

Program for converting .doc files from old Sharp PA-W1400 word processor to UTF-8 text.
Result should be importable in to word processor (e.g. OpenOffice/LibreOffice) with
character encoding UTF-8 and LF-only end of lines (typically auto-detected).

Compile with:

	gcc convert.c -Wall -o conv

Run as:

	./conv file.doc file.txt

Note it will not check/warn if file.txt already exists - just overwrites it.

Also note the 720k floppy used is likely to have a well-flaky file system and won't be
recognised by Linux or Windows from 98 onwards. See:

	http://support.microsoft.com/kb/140060/en-us

Best chance is to make an image of the floppy with something like:

	dd if=/dev/fd0 of=disk.img bs=512 count=1440 conv=noerror

Then copy the .img file and present it to a machine (real of virtual) running DOS. Use
the utility chkdsk to fix the file system (scandisk seems to bork) and then copy the
files off, probably using xcopy, as normal copy borked - possibly consequence of a
partly fixed but still flaky file system.

Copy to somewhere else running Linux to run this (or Windows, as this should compile OK
using cygwin or similar).

======================================================================================

Copyright Paul S. Crawford (c) 2011 all rights reserved

Email: psc 'at' sat.dundee.ac.uk

Licensed under GPL-2 (i.e. copy & use freely, return changes to society).

No warranty of any sort, user has to check the results are good enough. And I mean it, look
at the notes of how much I know (or don't) about the format.

25 Jul 2011 - psc Added guess for grave accented vowels.

=========================================================================================

Details as far as I could tell:

	Text appears to start at file offset 0x1B0, no idea what the header is for but use the
	first 4 bytes as a crude check that converting the file is a worth while thing to do.

Formatting of text:

	0x83 appears to be "bold character follows"

	Upper bit set implies underlined.

Escape sequence for odd characters:

	'0x81 0xB9' appear to be 'end of line / start of new line'?
	'0x81 0xC9' looks like tab? Spacing seems not quite right in simple text.
	'0x81 0xD6' appears to be tab, but only at line start?
	'0x81 0xBA' appear to be a space (line continue/break)?
	'0x81 0xBB' could be non-breaking hyphen, or hyphen at end of line?

	'0x8B {0x34 or others function/value unknown}' seems to be page feed / section break or similar?

	'0x10 0x00 0x75' seems to be u with umlaut (0x75 = 'u' in ASCII) for Zurich
	'0x11 0x00 0x5C' seems to be grave, not sure what 0x5C as probably just '1/2' character?
	'0x12 0x00 0x65' seems to be 'e' with acute
	No idea of similar sequence for circumflex (but suspect 0x09 or 0x13 then 0x00).

	'0x87 0x01' found once, also not sure what is is.
	'0x81 0xB8' found once at very start of document - tab or space? Not sure.

End of data:

	0x0A 0x78 appear to be end of text.
	0x8A 0x78 also seems to be end (0x0A with upper bit set).
	0x08 0x78 also seen as end of text (once).
	No idea what the stuff following the end of text is.

Single characters:

	0x0D seems to be a hyphen?
	0x0E could be hyphen, or maybe space?

	0xA0 => '_'   (underscore) for underscore (space + underline?)
	0x7D => '3/4' (normally }) for three quarters symbol
	0x5C => '1/2' (normally \) for one half symbol
	0x7F => '1/4' (normally del) for one quarter symbol
	0x1E => '£'   (record separator) for British pound

	But what (if anything) is used to actually print '{' or '\' then? Not seen on keyboard.

	Keyboard has a degree symbol, no idea of what code it has.

	Also it has 'c' with cedilla as in Français, no idea of code.

*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>


int main(int argc, char *argv[])
{
FILE *fpi, *fpo;
char *fnamei = argv[1];
char *fnameo = argv[2];
int c=0, flag=0, pc=0;
static int expected_hdr[] = {0xD3, 0xC8, 0xC1, 0xD2};
const int num_hdr = sizeof(expected_hdr)/sizeof(int);
int ii;

	//Check we have input & output specified.
	if(argc != 3)
		{
		fprintf(stderr, "Must have %s inputname outputname\n", argv[0]);
		exit(1);
		}

	//Open input for reading.
	fpi = fopen(fnamei, "rb");
	if(fpi == NULL)
		{
		fprintf(stderr, "Failed to open %s for reading (%s)\n", fnamei, strerror(errno));
		exit(1);
		}

	//Check is has the first bytes of the header (but is this version number dependent?)
	for(ii=0; ii<num_hdr; ii++)
		{
		c = fgetc(fpi);
		if(c != expected_hdr[ii])
			{
			fprintf(stderr, "Header of %s not as expected (0x%02X at %d != 0x%02X)\n", fnamei, c, ii, expected_hdr[ii]);
			exit(1);
			}
		}

	//Go to start of what appear to be the actual text. Don't know what the header is.
	if(fseek(fpi, 0x1B0, SEEK_SET))
		{
		fprintf(stderr, "Failed seeking to start of %s\n", fnamei);
		exit(1);
		}

	//Open output for writing.
	fpo = fopen(fnameo, "wb");
	if(fpo == NULL)
		{
		fprintf(stderr, "Failed to open %s for writing (%s)\n", fnameo, strerror(errno));
		exit(1);
		}

	//Tell the world what we age going to attempt.
	printf("Converting %s to %s\n", fnamei, fnameo);

	//Read input character-by-character, and attempt conversion process.
	while((c = fgetc(fpi)) != EOF)
		{
		if(c == 0x0A || c == 0x8A || c == 0x08)
			{
			if((c = fgetc(fpi)) == 0x78) break; //Appears to be end-of-text marker.
			}

		if(c == 0x83) continue; // Bold follows?

		if(c == 0x81)
			{
			flag = 1;
			continue; //Escape sequence with some other stuff (below)
			}

		if(c == 0x8B)
			{
			c = fgetc(fpi);
			printf("Page break? (0x8B 0x%02X sequence)\n", c);
			//fprintf(fpo, "(0x8B 0x%02X)", c);
			fputc('\n', fpo); //Just put in another line break - probably going to reformat anyway.
			continue;
			}

		if(flag)
			{
			//Escape sequence.
			switch(c)
				{
				case 0xA0: fputc('_', fpo);  break; //Underscore 0r hyphen?
				case 0xBA: fputc(' ', fpo);  break; //Space, but line-continue or similar?
				case 0xD6: fputc('\t', fpo); break; //Space/tab-like thing, start only?
				case 0xC9: fputc('\t', fpo); break; //Space/tab-like thing?
				case 0xB9: fputc('\n', fpo); break; //End of line / CR?
				case 0xBB: fputc('-', fpo);  break; //Hyphen or underscore thing?
				default:
					printf("Escape oddity 0x%02X\n", c);
					break;
				}
			}
		else
			{
			if(c == 0x10)
				{
				if((c = fgetc(fpi)) == 0x00)
					{
					// found 'umlaut' escape sequence?
					c = fgetc(fpi);
					if(c >= ' ' && c < 0x7F) printf("Found umlaut 0x%02X (%c)\n", c, c);

					// Change any recognised cases.
					if(c == 'a') c = 0x101; //guess - not seen
					if(c == 'e') c = 0x102; //guess - not seen
					if(c == 'i') c = 0x103; //guess - not seen
					if(c == 'o') c = 0x104; //guess - not seen
					if(c == 'u') c = 0x105; //u with umlaut, assign above 8-bit ASCII - so switch() handles exception.
					}
				}

			if(c == 0x11)
				{
				if((c = fgetc(fpi)) == 0x00)
					{
					// found 'grave' escape sequence?
					c = fgetc(fpi);
					if(c >= ' ' && c < 0x7F) printf("Found grave 0x%02X (%c)\n", c, c);

					if(c == 'a') c = 0x201; //guess - not seen
					if(c == 'e') c = 0x202; //guess - not seen
					if(c == 'i') c = 0x203; //guess - not seen
					if(c == 'o') c = 0x204; //guess - not seen
					if(c == 'u') c = 0x205; //guess - not seen
					}
				}

			if(c == 0x12)
				{
				if((c = fgetc(fpi)) == 0x00)
					{
					// found 'acute' escape sequence?
					c = fgetc(fpi);
					if(c >= ' ' && c < 0x7F) printf("Found acute 0x%02X (%c)\n", c, c);

					if(c == 'a') c = 0x301; //guess - not seen
					if(c == 'e') c = 0x302; //e with acute
					if(c == 'i') c = 0x303; //guess - not seen
					if(c == 'o') c = 0x304; //guess - not seen
					if(c == 'u') c = 0x305; //guess - not seen
					}
				}

			switch(c)
				{
				case 0x101: fputc(0xC3, fpo); fputc(0xA4, fpo); break; // 'a' with diaeresis (umlaut) in UTF-8
				case 0x102: fputc(0xC3, fpo); fputc(0xAB, fpo); break; // 'e' with diaeresis (umlaut) in UTF-8
				case 0x103: fputc(0xC3, fpo); fputc(0xAF, fpo); break; // 'i' with diaeresis (umlaut) in UTF-8
				case 0x104: fputc(0xC3, fpo); fputc(0xB4, fpo); break; // 'o' with diaeresis (umlaut) in UTF-8
				case 0x105: fputc(0xC3, fpo); fputc(0xBC, fpo); break; // 'u' with diaeresis (umlaut) in UTF-8

				case 0x201: fputc(0xC3, fpo); fputc(0xA0, fpo); break; // 'a' with grave UTF-8
				case 0x202: fputc(0xC3, fpo); fputc(0xA8, fpo); break; // 'e' with grave UTF-8
				case 0x203: fputc(0xC3, fpo); fputc(0xAC, fpo); break; // 'i' with grave UTF-8
				case 0x204: fputc(0xC3, fpo); fputc(0xB2, fpo); break; // 'o' with grave UTF-8
				case 0x205: fputc(0xC3, fpo); fputc(0xB9, fpo); break; // 'u' with grave UTF-8

				case 0x301: fputc(0xC3, fpo); fputc(0xA1, fpo); break; // 'a' with acute UTF-8
				case 0x302: fputc(0xC3, fpo); fputc(0xA9, fpo); break; // 'e' with acute UTF-8
				case 0x303: fputc(0xC3, fpo); fputc(0xAD, fpo); break; // 'i' with acute UTF-8
				case 0x304: fputc(0xC3, fpo); fputc(0xB3, fpo); break; // 'o' with acute UTF-8
				case 0x305: fputc(0xC3, fpo); fputc(0xBa, fpo); break; // 'u' with acute UTF-8


				case 0xA0: fputc('_', fpo); break; //Underscore or hyphen?

				case 0x8E:
				case 0x0E: fputc(' ', fpo); break; //Space or hyphen?

				case 0x8D:
				case 0x0D: fputc('-', fpo); break; //Hyphen or underscore thing? (also with upper bit set = underscore?)
#if 0
				//Ouput ASCII versions
				case 0x7D: fprintf(fpo, "%s", "(3/4)"/*"\u00be"*/); break; // '3/4' in UTF-8?
				case 0x5C: fprintf(fpo, "%s", "(1/2)"/*"\u00bd"*/); break; // '1/2' in UTF-8?
				case 0x7F: fprintf(fpo, "%s", "(1/4)"/*"\u00bc"*/); break; // '1/4' in UTF-8?
				case 0x1E: fprintf(fpo, "%s", "(#)"  /*"\u00a3"*/); break; // '£' in UTF-8?
#else
				//Output UTF-8 equivalents.
				case 0x7D: fputc(0xC2, fpo); fputc(0xBE, fpo); break; // '3/4' in UTF-8?
				case 0x5C: fputc(0xC2, fpo); fputc(0xBD, fpo); break; // '1/2' in UTF-8?
				case 0x7F: fputc(0xC2, fpo); fputc(0xBC, fpo); break; // '1/4' in UTF-8?

				case 0x9E:
				case 0x1E: fputc(0xC2, fpo); fputc(0xA3, fpo); break; // '£' in UTF-8? (and underscored version)
#endif
				default:
					//Strip upper bits, indicate underscore?
					pc = c & 0x7F;
					if(pc < 0x7F && pc >= ' ')
						{
						fputc(pc, fpo);
						}
					else
						{
						printf("Found unexpected character 0x%02X\n", c);
						}
					break;
				}
			}

		flag=0;
		}

	fclose(fpi);
	fclose(fpo);

	printf("Done\n");

return 0;
}
